! Copyright (C) 2023 Aleksander Sabak.
! See https://factorcode.org/license.txt for BSD license.
USING: accessors ascii assocs combinators continuations delegate
delegate.protocols hashtables io kernel lexer math math.parser
namespaces quotations sequences unicode vectors ;
IN: FALSE


ERROR: unsupported-FALSE-command command ;


DEFER: scan-FALSE


<PRIVATE


: peek-char ( -- char|f )
  lexer get dup still-parsing?
  [ dup still-parsing-line?
    [ [ column>> ] [ line-text>> ] bi nth ]
    [ drop CHAR: \n ] if ] [ drop f ] if ;

: consume-char ( -- )
  lexer get dup still-parsing-line?
  [ [ 1 + ] change-column drop ] [ next-line ] if ;

: scan-char ( -- char|f ) peek-char dup [ consume-char ] when ;


: (scan-FALSE-number) ( digit -- number )
  CHAR: 0 - [ peek-char dup ascii:digit? ]
  [ consume-char CHAR: 0 - swap 10 * + ] while drop ;


: (;) ( state -- state )
  dup pop over at* [ drop 0 ] unless suffix! ;

: (:) ( state -- state )
  dup [ pop ] [ pop ] bi swap pick set-at ;

: ($) ( state -- state ) dup last suffix! ;

: (%) ( state -- state ) dup pop* ;

: (\) ( state -- state )
  dup [ pop ] [ pop ] bi [ suffix! ] dip suffix! ;

: (@) ( state -- state )
  dup [ pop ] [ pop ] [ pop ] tri
  [ swap [ suffix! ] dip suffix! ] dip suffix! ;

: (O) ( state -- state )
  dup dup [ length ] [ pop ] bi - 2 - swap nth suffix! ;

: (+) ( state -- state ) dup [ pop ] [ pop ] bi + suffix! ;

: (-) ( state -- state ) dup [ pop ] [ pop ] bi swap - suffix! ;

: (*) ( state -- state ) dup [ pop ] [ pop ] bi * suffix! ;

: (/) ( state -- state )
  dup [ pop ] [ pop ] bi swap /i suffix! ;

: (_) ( state -- state ) dup pop neg suffix! ;

: (&) ( state -- state ) dup [ pop ] [ pop ] bi bitand suffix! ;

: (|) ( state -- state ) dup [ pop ] [ pop ] bi bitor suffix! ;

: (~) ( state -- state ) dup pop bitnot suffix! ;

: (=) ( state -- state )
  dup [ pop ] [ pop ] bi = [ -1 ] [ 0 ] if suffix! ;

: (>) ( state -- state )
  dup [ pop ] [ pop ] bi < [ -1 ] [ 0 ] if suffix! ;

: (!) ( state -- state ) dup pop call( state -- state ) ;

: (?) ( state -- state )
  dup [ pop ] [ pop ] bi 0 =
  [ drop ] [ call( state -- state ) ] if ;

: (#) ( state -- state )
  dup [ pop [ call( state -- state ) ] curry ]
  [ pop [ call( state -- state ) dup pop 0 = ] curry ] bi
  swap until ;

: (^) ( state -- state ) read1 -1 or suffix! ;

: (,) ( state -- state ) dup pop write1 ;

: (.) ( state -- state ) dup pop number>string write ;

ALIAS: (B) flush


CONSTANT: FALSE-dictionary
  H{
    { CHAR: ; (;) }
    { CHAR: : (:) }
    { CHAR: $ ($) }
    { CHAR: % (%) }
    { CHAR: \ (\) }
    { CHAR: @ (@) }
    { CHAR: O (O) }
    { CHAR: ø (O) }
    { CHAR: + (+) }
    { CHAR: - (-) }
    { CHAR: * (*) }
    { CHAR: / (/) }
    { CHAR: _ (_) }
    { CHAR: & (&) }
    { CHAR: | (|) }
    { CHAR: ~ (~) }
    { CHAR: = (=) }
    { CHAR: > (>) }
    { CHAR: ! (!) }
    { CHAR: ? (?) }
    { CHAR: # (#) }
    { CHAR: ^ (^) }
    { CHAR: , (,) }
    { CHAR: . (.) }
    { CHAR: B (B) }
    { CHAR: ß (B) }
  }


CONSTANT: FALSE-sigils
  H{
    { CHAR: {
      [ [ scan-char
          { { CHAR: } [ f ] }
            { f [ \ } throw-unexpected-eof ] }
            [ drop t ] } case ] loop ] }
    { CHAR: "
      [ V{ } clone
        [ scan-char
          { { CHAR: " [ f ] }
            { f [ \ " throw-unexpected-eof ] }
            [ suffix! t ] } case ] loop suffix!
        \ write suffix! ] }
    { CHAR: ' [ scan-char suffix! \ suffix! suffix! ] }
    { CHAR: [ [ scan-FALSE suffix! \ suffix! suffix! ] }
    { CHAR: ` [ CHAR: ` unsupported-FALSE-command ] }
  }


PRIVATE>


! FALSE state

TUPLE: FALSE-state { stack vector } { dictionary hashtable } ;

: <FALSE-state> ( stack dictionary -- state )
  [ V{ } clone-like ] [ H{ } assoc-clone-like ] bi*
  FALSE-state boa ;

CONSULT: assoc-protocol FALSE-state dictionary>> ;

CONSULT: sequence-protocol FALSE-state stack>> ;

M: FALSE-state like drop { } <FALSE-state> ;

M: FALSE-state new-sequence
  state>> new-sequence { } <FALSE-state> ;

M: FALSE-state new-resizable
  state>> new-resizable { } <FALSE-state> ;

M: FALSE-state new-assoc
  dictionary>> new-assoc { } swap <FALSE-state> ;

M: FALSE-state assoc-like
  dictionary>> assoc-like { } swap <FALSE-state> ;

INSTANCE: FALSE-state sequence

INSTANCE: FALSE-state assoc

: <empty-FALSE-state> ( -- state ) { } { } <FALSE-state> ;


: with-FALSE
  ( ..A seq q: ( ..A state -- ..B state' ) -- ..B seq' )
  swap [ H{ } <FALSE-state> swap call stack>> ] keep
  clone-like ; inline


: scan-FALSE ( -- quote )
  V{ } clone
  [ scan-char dup CHAR: ] = ]
  [ { { [ FALSE-dictionary ?at ] [ suffix! ] }
      { [ FALSE-sigils ?at ] [ call( x -- x ) ] }
      { [ dup ascii:digit? ]
        [ (scan-FALSE-number) suffix! \ suffix! suffix! ] }
      { [ dup unicode:blank? ] [ drop ] }
      [ suffix! \ suffix! suffix! ] } cond ] until
  drop >quotation ;


SYNTAX: FALSE[ scan-FALSE suffix! ;
